import java.awt.*;
import java.applet.*;
/*
        Project Java, LICIL
	Author: Konstantin Lukin 
        (http://www.eserc.stonybrook.edu/ProjectJava/WaveInt/)
        klukin77@yahoo.com
*/        
public class Interaction extends Applet implements Runnable {

private Rectangle border, graph, clip, control, space[], base[], part[];
private SineWave wave[], store_wave[];
private SumWaves sum;
private Box box[], speed_box[];
private Handle[] handle;
private Scrollbar bar[], speed[];
private Image[] image;
private Graphics[] offg;
private FontMetrics fm;
private Font label_font, graph_font, title_font;
private int index, update_mode, values_mode, runner_state, speed_value[];
private Thread runner;
Button button;

public void init() {
   title_font = new Font("Helvetica", Font.BOLD , 16);
   label_font = new Font("Helvetica", Font.PLAIN, 14);
   graph_font = new Font("Helvetica", Font.PLAIN, 12);
   speed_value = new int[2];
   for(int i=0; i<speed_value.length; i++) speed_value[i] = 5;
   setLayout(null);
   defineLayout();   
   setWaves();
   setComponents();
   setHandles();
   index = -1;
   update_mode = 2;
   values_mode = 6;
   runner_state = 0;
}

public String getAppletInfo() {
   return "Waveform Interaction, written by Konstantin Lukin. (ProjectJava, LICIL)";
}


//---------------------------------------------------------------
public void start() {
   runner = new Thread(this);
   runner.setPriority(Thread.MIN_PRIORITY);
   runner.start();
}

public void stop() {
   runner.stop();
   runner = null;
}

public void run() {
   while(true) {
      pause(30);
      runner.yield();
      switch(runner_state) {
         case 0:
            break;
         case 1:
            wave[0].advance(speed_value[0]);
            wave[1].advance(speed_value[1]);
            repaint();            
            break;
         default:
      }
   }
}

private void pause(int time) {
   try { runner.sleep(time); }
   catch(InterruptedException e) {}
}

private void startAnimation() {
   store_wave[0] = wave[0].copy();
   store_wave[1] = wave[1].copy();
   for(int i=0; i<2; i++) bar[3*i+2].disable();  
   update_mode = 2;
   values_mode = -2;
   runner_state = 1;
}

private void stopAnimation() {
   runner_state = 0;
   for(int i=0; i<2; i++) wave[i].phase = store_wave[i].phase;
   for(int i=0; i<2; i++) bar[3*i+2].enable();
   update_mode = 2;
   values_mode = 6;
   repaint();
}
//---------------------------------------------------------------
public boolean action(Event evt, Object arg) {
   if(evt.target instanceof Button) {
      if(button.getLabel().equals("start")) {
         startAnimation();
         button.setLabel("stop");
      }
      else {
         stopAnimation();
         button.setLabel("start");
      }
   }
   else return super.action(evt, arg);
   return true;
}

//---------------------------------------------------------------
public boolean handleEvent(Event evt) {
   if(evt.id == Event.SCROLL_ABSOLUTE  || evt.id == Event.SCROLL_LINE_UP   || 
      evt.id == Event.SCROLL_LINE_DOWN || evt.id == Event.SCROLL_PAGE_UP   || 
                                          evt.id == Event.SCROLL_PAGE_DOWN)  {
      int i;
      for(i=0; i<8; i++)  {
         if(i<6 && evt.target == bar[i])     break;
         else if(i>5 && evt.target == speed[i-6]) break;
      }
      switch(i) {
         case 0: case 3:
            wave[i/3].wavelength = bar[i].getValue() * 20/9;
            values_mode = (i==0) ? 0 : 2;
            break;
         case 1: case 4:
            wave[i/3].amplitude  = bar[i].getValue() * 1/2;
            values_mode = (i==1) ? 0 : 2; 
            break;
         case 2: case 5:
            wave[i/3].phase      = bar[i].getValue() * 4;
            values_mode = (i==2) ? 1 : 3;
            break;
         case 6: case 7:
            speed_value[i-6] = (speed[i-6].getValue()/3);
            values_mode = (i==6) ? 4 : 5;
            break;
         default:
      }
      update_mode = i/3;
   }
   else return super.handleEvent(evt);
   repaint();
   return true;
}


//---------------------------------------------------------------
public boolean mouseDown(Event evt, int x, int y) {
   if(runner_state==1) return super.mouseDown(evt,x,y);
   index = insideHandle(x,y);
   if(index==-1) return super.mouseDown(evt, x, y);
   else {
      notifyBars(Color.lightGray);
   }
   return true;
}

public boolean mouseUp(Event evt, int x, int y) {
   if(runner_state==1) return super.mouseDown(evt,x,y);
   notifyBars(Color.gray);
   index = -1;
   return super.mouseUp(evt, x, y);
}

public boolean mouseDrag(Event evt, int x, int y) {
   int i, newx, newy;
   switch(index) {
      case -1:
         return super.mouseDrag(evt, x, y);
      case 0: case 2:
         i = index/2;
         newx = x - space[i].x - base[i].x;
         int phase = AdjustPhase(newx*360/wave[i].wavelength);
         wave[i].phase = phase;
         updateBars();
         update_mode = (index==0) ? 0 : 1;
         values_mode = (index==0) ? 1 : 3;
         break;
      case 1: case 3:
         i = index/2;
         newx = x - space[i].x - base[i].x;
         newy = y - space[i].y - base[i].y;
         int amp = AdjustAmp(base[i].height/2 - newy);
         wave[i].amplitude = amp;
         int lambda = (newx*360)/(wave[i].phase+90);
         lambda = AdjustLambda(lambda);
         wave[i].wavelength = lambda;
         updateBars();
         update_mode = (index==1) ? 0 : 1;
         values_mode = (index==1) ? 0 : 2;
         break;
      default:
         return super.mouseDrag(evt, x, y); 
   }
   repaint();
   return true;
}

private int AdjustLambda(int lambda) {
   if(lambda > 200) return 200;
   else if(lambda <5) return 5;
   else return lambda;
}

private int AdjustAmp(int amp) {
   if(amp > 45) return 45;
   else if(amp <0) return 0;
   else return amp;
}

private int AdjustPhase(int phase) {
   if(phase >360) return 360;
   else if(phase <0) return 0;
   else return phase;
}

private int insideHandle(int x, int y) {
   int newx, newy;
   for(int i=0; i<handle.length; i++) {
      newx = x - space[i/2].x;
      newy = y - space[i/2].y;
      if(handle[i].inside(newx, newy)) return i;
   }
   return -1;
}

private void notifyBars(Color color) {
   switch(index) {
      case -1:
         return;
      case 0:
         bar[2].setBackground(color);       
         break;
      case 1:
         bar[0].setBackground(color);
         bar[1].setBackground(color);
         break;
      case 2:
         bar[5].setBackground(color);
         break;
      case 3:
         bar[3].setBackground(color);     
         bar[4].setBackground(color);
         break;
   }
}

private void updateBars() {
   switch(index) {
      case -1:
         return;
      case 0:
         bar[2].setValue(wave[0].phase/4);       
         break;
      case 1:
         bar[0].setValue(wave[0].wavelength*9/20);
         bar[1].setValue(wave[0].amplitude*2);
         break;
      case 2:
         bar[5].setValue(wave[1].phase/4);
         break;
      case 3:
         bar[3].setValue(wave[1].wavelength*9/20);
         bar[4].setValue(wave[1].amplitude*9/4);
         break;
   }
}
//---------------------------------------------------------------
public void paint(Graphics g) {
   if(offg == null) {
      image = new Image[3];
      offg  = new Graphics[3];
      for(int i=0; i<image.length; i++) {
         image[i] = createImage(space[i].width, space[i].height);
         offg[i]  = image[i].getGraphics();
      }
   }
   drawBorder(g);
   drawControl(g);
   drawGraph(g);
   update_mode = 2;
   values_mode = 6;
   update(g);
}

public void update(Graphics g) {
   drawValues(g);
   if(runner_state==1) { update_mode = 2; values_mode = -2; }
   switch(update_mode) {
      case -1:
         return;
      case 0:
         drawImage0();
	 drawImage2();
         g.drawImage(image[0], space[0].x, space[0].y, null);
         g.drawImage(image[2], space[2].x, space[2].y, null); 
         break;
      case 1:
         drawImage1();
	 drawImage2();
         g.drawImage(image[1], space[1].x, space[1].y, null);
         g.drawImage(image[2], space[2].x, space[2].y, null); 
         break;
      case 2:
         drawImage0();
         drawImage1();
	 drawImage2();
         g.drawImage(image[0], space[0].x, space[0].y, null);
         g.drawImage(image[1], space[1].x, space[1].y, null); 
         g.drawImage(image[2], space[2].x, space[2].y, null); 
         break;
      default:
         return;
   }
}

/*--------------------------------------------------------------
----------------------------------------------------------------*/
private void setHandles(int i) {
   handle[2*i].move(base[i].x + (wave[i].phase*wave[i].wavelength)/360, space[i].height/2);
   handle[2*i+1].move(handle[2*i].Cx+wave[i].wavelength/4, space[i].height/2-wave[i].amplitude);
}

void drawGraphPaper(int mode, int start1, int start2) {
   offg[mode].setColor(Color.darkGray);
   for(int i=start1; i<space[mode].width; i+=10) {
      offg[mode].drawLine(i, 0, i, space[mode].height);
   }
   for(int i=start2; i<space[mode].height; i+=10) {
      offg[mode].drawLine(0, i, space[mode].width, i);
   }
}

//---------------
private void drawImage0() {
   offg[0].setColor(Color.black);
   offg[0].fillRect(0,0,space[0].width, space[0].height);
   drawGraphPaper(0, 0, 8);
   offg[0].setColor(new Color(0, 180, 0));
   wave[0].draw(offg[0],  base[0].x, space[0].height/2, base[0].width, 2);
   if(runner_state==0) {
      setHandles(0);
      offg[0].setColor(Color.white);
      handle[0].draw(offg[0]);
      handle[1].draw(offg[0]);
   }
}

//---------------------
private void drawImage1() {
   offg[1].setColor(Color.black);
   offg[1].fillRect(0,0,space[1].width, space[1].height);
   drawGraphPaper(1, 0, 4);
   offg[1].setColor(new Color(0, 180, 0));
   wave[1].draw(offg[1],  base[1].x, space[1].height/2, base[1].width, 2);
   if(runner_state == 0) {
      setHandles(1);
      offg[1].setColor(Color.white);
      handle[2].draw(offg[1]);
      handle[3].draw(offg[1]);
   }
}

//----------------------------
private void drawImage2() {
   offg[2].setColor(Color.black);
   offg[2].fillRect(0,0,space[2].width, space[2].height);
   drawGraphPaper(2, 0, 0);
   offg[2].setColor(Color.blue);
   sum.draw(offg[2],  base[2].x, space[2].height/2, base[2].width, 3);
}

//---------------------------------------------------------------
void drawBorder(Graphics g) {
   g.setColor(Color.black);
   g.fillRect(border.x, border.y, border.width, border.height);
   g.setColor(Color.white);
   g.drawRect(border.x-1, border.y-1, border.width, border.height);
   g.setColor(Color.gray);
   g.drawRect(border.x+1, border.y+1, border.width, border.height);
   g.setColor(Color.black);
   g.drawRect(border.x, border.y, border.width, border.height);
}


//---------------------------------------------------------------
void drawGraph(Graphics g) {
   g.setColor(Color.black);
   g.fillRect(graph.x, graph.y, graph.width+1, graph.height+1);

   drawGraphPaper(g);
   drawGraphLabels(g);
}

void drawGraphPaper(Graphics g) {
   g.setColor(Color.darkGray);
   for(int i=graph.x; i<graph.x+graph.width; i+=10) {
      g.drawLine(i, graph.y, i, graph.y+graph.height);
   }
   for(int i=graph.y; i<graph.y+graph.height; i+=10) {
      g.drawLine(graph.x, i, graph.x+graph.width, i);
   }
}

void drawGraphLabels(Graphics g) {
   g.setFont(graph_font);
   g.setColor(Color.lightGray);
   g.drawString("Wave1", graph.x+5, space[0].y+space[0].height/2); 
   g.drawString("Wave2", graph.x+5, space[1].y+space[1].height/2);
   g.drawString("Sum"  , graph.x+5, space[2].y+space[2].height/2);
}


//---------------------------------------------------------------
void drawControl(Graphics g) {
   g.setColor(Color.black);
   g.fillRect(control.x, control.y, control.width, control.height);
   drawWaveBox0(g);
   drawWaveBox1(g);
   drawAnimBox(g);
   drawSeparator(g);   
   drawControlLabels(g);
}

void drawControlLabels(Graphics g) {
   g.setFont(label_font);
   String[] label = { "Wavelength", "Amplitude", "Phase" };
   for(int i=0; i<6; i++) {
      g.setColor(Color.black);
      box[i].fill(g, 0);
      g.setColor(Color.green);
      g.drawString(label[i%3], box[i].x, box[i].y+box[i].top.height-5);
   }
   String[] label1 = { "Speed 1", "Speed 2" };
   for(int i=0; i<2; i++) {
      g.setColor(Color.black);
      speed_box[i].fill(g, 0);
      g.setColor(Color.green);
      g.drawString(label1[i], speed_box[i].x, speed_box[i].y+speed_box[i].top.height-5);
   }
   label = null;
   label1=null;
}

private void drawValues(Graphics g) {
   g.setFont(label_font);
   switch(values_mode) {
      case -2:
         break;
      case -1:
         g.setColor(Color.black);
         for(int i=0; i<6; i++) box[i].fill(g, 4);
         for(int i=0; i<2; i++) speed_box[i].fill(g, 4);
         break;
      case 0:
         g.setColor(Color.black);
         box[0].fill(g, 4);
         box[1].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+wave[0].wavelength, box[0].right_top.x, box[0].y+box[0].top.height-5); 
         g.drawString(""+wave[0].amplitude , box[1].right_top.x, box[1].y+box[1].top.height-5);
         break;
      case 1:
         g.setColor(Color.black);
         box[2].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+wave[0].phase     , box[2].right_top.x, box[2].y+box[2].top.height-5);
         break;
      case 2:
         g.setColor(Color.black);
         box[3].fill(g, 4);
         box[4].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+wave[1].wavelength, box[3].right_top.x, box[3].y+box[3].top.height-5); 
         g.drawString(""+wave[1].amplitude , box[4].right_top.x, box[4].y+box[4].top.height-5);
         break;
      case 3:
         g.setColor(Color.black);
         box[5].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+wave[1].phase     , box[5].right_top.x, box[5].y+box[5].top.height-5);
         break;
      case 4:
         g.setColor(Color.black);
         speed_box[0].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+speed_value[0], speed_box[0].right_top.x, 
                         speed_box[0].y+speed_box[0].top.height-5);
         break;
      case 5:
         g.setColor(Color.black);
         speed_box[1].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+speed_value[1], speed_box[1].right_top.x, 
                         speed_box[1].y+speed_box[1].top.height-5);
         break;
      case 6:
         g.setColor(Color.black);
         for(int i=0; i<6; i++) box[i].fill(g, 4);
         for(int i=0; i<2; i++) speed_box[i].fill(g, 4);
         g.setColor(Color.green);
         g.drawString(""+wave[0].wavelength, box[0].right_top.x, box[0].y+box[0].top.height-5); 
         g.drawString(""+wave[0].amplitude , box[1].right_top.x, box[1].y+box[1].top.height-5);
         g.drawString(""+wave[0].phase     , box[2].right_top.x, box[2].y+box[2].top.height-5);
         g.drawString(""+wave[1].wavelength, box[3].right_top.x, box[3].y+box[3].top.height-5); 
         g.drawString(""+wave[1].amplitude , box[4].right_top.x, box[4].y+box[4].top.height-5);
         g.drawString(""+wave[1].phase     , box[5].right_top.x, box[5].y+box[5].top.height-5); 
         g.drawString(""+speed_value[0], speed_box[0].right_top.x, 
                         speed_box[0].y+speed_box[0].top.height-5);
         g.drawString(""+speed_value[1], speed_box[1].right_top.x, 
                         speed_box[1].y+speed_box[1].top.height-5);
         break;
      default:
   }
}

private void drawSeparator(Graphics g) {
   g.setColor(Color.lightGray);
   g.drawLine(control.x, control.y, control.x, control.y+control.height);
}

private void drawWaveBox0(Graphics g) {
   Rectangle r; String s;
   fm = g.getFontMetrics(title_font);   
   s = "Wave 1";

   r = new Rectangle(part[0].x+5, part[0].y+10, part[0].width-10, part[0].height-15);
   g.setColor(Color.lightGray);
   g.drawRoundRect(r.x, r.y, r.width, r.height, 20, 20);

   int w = fm.stringWidth(s)+10;
   r = new Rectangle(r.x+(r.width-w)/2, r.y-10, w, 15);
   g.setColor(Color.black);
   g.fillRect(r.x, r.y, r.width, r.height);   

   g.setFont(title_font);
   g.setColor(Color.red);
   g.drawString(s, r.x+(r.width-fm.stringWidth(s))/2, r.y+r.height);
   fm = null;
}

private void drawWaveBox1(Graphics g) {
   Rectangle r; String s;
   fm = g.getFontMetrics(title_font);
   s = "Wave 2";

   r = new Rectangle(part[1].x+5, part[1].y+10, part[1].width-10, part[1].height-15);
   g.setColor(Color.lightGray);
   g.drawRoundRect(r.x, r.y, r.width, r.height, 20, 20);

   int w = fm.stringWidth(s)+10;
   r = new Rectangle(r.x+(r.width-w)/2, r.y-10, w, 15);
   g.setColor(Color.black);
   g.fillRect(r.x, r.y, r.width, r.height);   

   g.setFont(title_font);
   g.setColor(Color.red);
   g.drawString(s, r.x+(r.width-fm.stringWidth(s))/2, r.y+r.height);
   fm = null;
}


private void drawAnimBox(Graphics g) {
   Rectangle r; String text;
   g.setFont(title_font);
   fm = g.getFontMetrics(title_font);
   text = "Animation";

   r = new Rectangle(part[2].x+5, part[2].y+10, part[2].width-10, part[2].height-15);
   g.setColor(Color.lightGray);
   g.drawRoundRect(r.x, r.y, r.width, r.height, 20, 20);

   int w = fm.stringWidth(text)+10;
   r = new Rectangle(r.x+(r.width-w)/2, r.y-10, w, 15);
   g.setColor(Color.black);
   g.fillRect(r.x, r.y, r.width, r.height);   

   g.setColor(Color.red);
   g.drawString(text, r.x+(r.width-fm.stringWidth(text))/2, r.y+r.height);
   fm = null;
}

//---------------------------------------------------------------
private void setHandles() {
   handle = new Handle[4];
   for(int i=0; i<handle.length; i++) {
      if(i%2==0) {
         handle[i] = new Handle(Handle.HORIZONTAL);
         handle[i].move(base[i/2].x, base[i/2].y+base[i/2].height/2);
      }
      else {
         handle[i] = new Handle(Handle.BOTH);
         handle[i].move(base[i/2].x+wave[i/2].wavelength/4, 
                        base[i/2].y+base[i/2].height/2-wave[i/2].amplitude);
      }
   }
}


private void setComponents() {
   box       = new Box[6];
   speed_box = new Box[2];
   bar       = new Scrollbar[6];
   speed     = new Scrollbar[2];

   for(int i=0; i<2; i++) {
      int j = 3*i;
      bar[j] = new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, 90);
      bar[j+1] = new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, 90);
      bar[j+2] = new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, 90);
      speed[i] = new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, 90);

      add(bar[j]);
      add(bar[j+1]);
      add(bar[j+2]);
      add(speed[i]);

      bar[j].setBackground(Color.gray);
      bar[j+1].setBackground(Color.gray);
      bar[j+2].setBackground(Color.gray);
      speed[i].setBackground(Color.gray);

      bar[j].setValue(wave[i].wavelength  * 9/20);
      bar[j+1].setValue(wave[i].amplitude * 2);
      bar[j+2].setValue(wave[i].phase     * 1/4); 
      speed[i].setValue(speed_value[i]*3);
   }

   Rectangle r;
   r = new Rectangle(part[0].x+10, part[0].y+20, part[0].width-20, part[0].height-30);
   for(int i=0; i<3; i++) {
      box[i] = new Box(r.x, r.y+i*r.height/3, r.width, r.height/3);
      bar[i].reshape(box[i].x, box[i].left.y, box[i].width, box[i].left.height);
   }

   r = new Rectangle(part[1].x+10, part[1].y+20, part[1].width-20, part[1].height-30);
   for(int i=3; i<6; i++) {
      box[i] = new Box(r.x, r.y+(i-3)*r.height/3, r.width, r.height/3);
      bar[i].reshape(box[i].x, box[i].left.y, box[i].width, box[i].left.height);
   }

   r = new Rectangle(part[2].x+10, part[2].y+20, part[2].width-20, part[2].height-30);   
   for(int i=0; i<speed.length; i++) {
      speed_box[i] = new Box(r.x, r.y+i*r.height/3, r.width, r.height/3);
      speed[i].reshape(speed_box[i].x, speed_box[i].left.y, 
                       speed_box[i].width, speed_box[i].left.height);
   }
   button = new Button("start");
   button.setFont(title_font);
   button.setBackground(Color.lightGray);
   button.setForeground(Color.black);
   add(button);
   button.reshape(r.x+20, r.y+2*r.height/3+3, r.width-40, r.height/3-6);
}


private void setWaves() {
   wave = new SineWave[2];
   store_wave = new SineWave[2];
   wave[0] = new SineWave(40, 40, 0);
   wave[1] = new SineWave(30, 40, 0);
   sum     = new SumWaves(wave[0], wave[1]);
}


private void defineLayout() {
   border = new Rectangle(5, 5, size().width-10, size().height-10); 
   graph  = new Rectangle(border.x+5, border.y+5, border.width-150, border.height-10);   
   clip   = new Rectangle(graph.x+50, graph.y+2, graph.width-60, graph.height-4);
//   msgbox = new Rectangle(graph.x, border.y+border.height-40, graph.width, 35);
   control= new Rectangle(graph.x+graph.width+5, border.y+5, 138, border.y+border.height-15);
   
   space = new Rectangle[3]; // bounds of offscreen images with absolute coordinates
   space[0] = new Rectangle(clip.x, clip.y,               clip.width, clip.height/4);
   space[1] = new Rectangle(clip.x, clip.y+clip.height/4, clip.width, clip.height/4);
   space[2] = new Rectangle(clip.x, clip.y+clip.height/2, clip.width, clip.height/2);

   base = new Rectangle[3];  // bounds of waves with coordinates relative to space[]
   base[0] = new Rectangle(10, (space[0].height-90)/2, space[0].width-20, 90);
   base[1] = new Rectangle(10, (space[1].height-90)/2, space[1].width-20, 90);
   base[2] = new Rectangle(10, (space[2].height-180)/2, space[2].width-20, 180);

   part = new Rectangle[3];  // control panel subdivision
   part[0] = new Rectangle(control.x, control.y, control.width, control.height*7/20);
   part[1] = new Rectangle(control.x, part[0].y+part[0].height, control.width, control.height*7/20);
   part[2] = new Rectangle(control.x, part[1].y+part[1].height, control.width, control.height*6/20);
}
}


   
   
   
